#include "mbed.h"
#include "math.h"

#define DEVELOP //ここをコメントアウトすることでシリアル通信を停止できる
#ifdef DEVELOP
#define dbg(...) pc.printf(__VA_ARGS__)
#else
#define dbg(...)
#endif


#define SAMPLE_RATE 5120

PwmOut ledR(D2);
PwmOut ledG(D3);
PwmOut ledB(D4);

AnalogIn guitar(A0);
Serial pc(USBTX, USBRX);
DigitalIn irSensor(D10);
DigitalOut irLed(D11);
DigitalIn sw1(D12);


// --- State ---
enum State_enum {
    START, WAIT, MEMORIZE, RECORD, CHECK, TEST
};
uint8_t state = START; // 最初はSTARTから
bool memorize_flag = 0; // WAIT -> MEMORIZEに移行するかどうか

// --- IR send ---
#define  REMOCON_CODE_LEN 400
#define  REMOCON_CAPTURE_INTERVAL_US  50
#define  IR_TRANSMIT_CARRIER_HZ 38000

uint8_t data[REMOCON_CODE_LEN];
Ticker irTick;


// --- FFT ---
#define FFTSIZE 512 // n >= 2, n = power of 2
const int IP_LEN = 4 + sqrt((double)(FFTSIZE/2)); // FFTライブラリ導入用:length of ip >= 2+sqrt(n/2)
const int CAND_LEN = 3; // 信号周波数の候補を出す数
const int PHRASE_LEN = 1000; // フレーズ記憶配列の最大長
const float SEC_PER_SAMPLEPOINT = 1.0 / SAMPLE_RATE; // length of ip >= 2+sqrt(n/2)
const float FFT_RESOLUTION = SAMPLE_RATE/FFTSIZE; // FFTの分解能

// --- Sampling関連 ---
Ticker tick; // 繰り返しタイマー割り込みを宣言
double *smpBufp = NULL; // サンプリング用バッファのポインタ
int countTick; // Ticker起動からの繰り返し回数
Timer timer; // タイマーを宣言
bool smpDone; //record完了フラグ


// --- Recording ---
//int *phrase = NULL;
int phrase[PHRASE_LEN];
int p_cnt = 0;

int summary[2][100]; // 録音したフレーズを1行目 周波数, 2行目 同一音の長さ で記録
int s_cnt=0; 
int soundLength = 0;
int memory[2][100];    
int m_len; // memoryに音が記録されている数

Timeout ledTick;

// --- Functions ---
void rdft(int n, int isgn, double *a, int *ip, double *w); // fftメソッド aにx(t)配列を入力すると、X(f)で更新される

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

  // サンプリング処理
  void flip() {
    // Bufferにanalog readした値を詰め込む
    smpBufp[countTick++] = guitar.read() - 0.5; // AnalogIn.read() の値範囲は 0～1.0 ｎなので、中心値を 0 に
    if (countTick == FFTSIZE) { // FFTで処理するサイズになったら
      tick.detach(); // tickerを取り除く
      smpDone = true;
    }
  }

// 大浦ライブラリでFFTを実行する (output: 0なら検出なし、他の数値なら音を)
int doFFT(double *FFTBuffer) { 
  int ip[IP_LEN]; // FFT関数実行用
  double mean = 0; // マグニチュードのmean
  double sd = 0; // マグニチュードのsd
  
  // -------- FFT
  double *w = (double*)malloc(sizeof(double) * (FFTSIZE/2)); //FFT実行用 work area
  
  double amp[FFTSIZE/2];
  int cand_cnt = 0;
  
  ip[0] = 0;
  
  rdft(FFTSIZE, 1, FFTBuffer, ip, w);

  double maxAmp[CAND_LEN]; 
  int frq[CAND_LEN];
  for (int i = 0; i < FFTSIZE/2; i++) {
    double a = FFTBuffer[i*2]; // 実部
    double b = FFTBuffer[i*2 + 1]; // 虚部
    double r = sqrt(a*a + b*b); // a+ib の絶対値 √ a^2 + b^2 = r が振幅値
    amp[i] = r;
  }

  // -------- 閾値を算出する  
  for (int i = 0; i<FFTSIZE/2 ; i++) {
    mean += amp[i];         
  }
  mean = mean/(FFTSIZE/2);

  for (int i = 0; i<FFTSIZE/2; i++) {
    sd += (amp[i]-mean) * (amp[i]-mean);
  }
  sd = sqrt(sd/(FFTSIZE/2));
  
  double threshold = mean + 3.0*sd;
  
  // -------- 周波数の候補値を下から順に出す
  for (int i = 0; i<FFTSIZE/2; i++) {
     if( cand_cnt < CAND_LEN && amp[i] > threshold && i * FFT_RESOLUTION > 60 ) {
      maxAmp[cand_cnt] = amp[i];
      frq[cand_cnt] = i * FFT_RESOLUTION;
      cand_cnt++;
    }  
  }
  
  free(w);

  //  for (int i=0; i<cand_cnt; i++){
  //    dbg("peek%d freq=%d mag=%f\r\n", i, (int)frq[i], maxAmp[i]); 
  //  }
  
  if(cand_cnt>0) {
    // dbg("freq=%d mag=%f\r\n", (int)frq[0], maxAmp[0]);
    return (int)frq[0];
  } else {
    return 0;    
  }
}

int detectFreq() {
    countTick = 0; // 周期処理用のカウントを初期化
    smpDone = false;

    tick.attach(&flip, SEC_PER_SAMPLEPOINT); // 指定したサンプリング間隔でflipを実行

    while (!smpDone) { // サンプリングが終わるまで待つ
      wait_ms(100);
    }
    int freq = doFFT(smpBufp); // fft実行

    return freq;
}

// state遷移処理
void toWait() {
    memorize_flag = 0;
    
    memset(phrase, 0, sizeof(phrase));
    //phrase[PHRASE_LEN] = {};
    p_cnt = 0;
    s_cnt = 0;
    soundLength = 0;
    
    for(int i=0; i<100; i++) {
        summary[0][i] = 0;
        summary[1][i] = 0;
    }

    dbg("WAIT\r\n");  
    state = WAIT;
}




bool captureIrCommand(uint8_t* data)
{
  memset(data,0,REMOCON_CODE_LEN);
  while(irSensor){} //赤外線受光器がLOWになったら抜ける(赤外線を受信したら）
  float uTime = timer.read_us(); //micro秒単位で赤外線を受信した時刻を記録
  for(int count=0;count<REMOCON_CODE_LEN;count++){
    for(int i=0;i<8;i++){
      data[count] |= irSensor << i;
      uTime += REMOCON_CAPTURE_INTERVAL_US;
      while(timer.read_us() < uTime) {
        //float t =timer.read_us();
        //dbg("wait %f \r\n",t );
      }
    }
  }
  

  dbg(" ---- Received ---- \r\n");
  for(int i=0;i<REMOCON_CODE_LEN;i++){
    dbg("%d ",data[i]);
    if(i != 0 && i % 16 == 0){
      dbg("\r\n");
    }
  }
  dbg(" ------------------- \r\n");
  
  
  
  return 1;
}


void transmitIrCommand(uint8_t* data)
{
  
  dbg("Transmit IR Command \r\n");
  //以下のforループはシリアルモニタに送信中の赤外線信号を表示するもの
  
  float uTime = timer.read_us();
  for(int count=0;count<REMOCON_CODE_LEN;count++){
    for(int i=0;i<8;i++){
      if((data[count] >> i) & 0x01){
        irLed = 0;
        //pinMode(PIN_IR_OUT,INPUT);
      }
      else {
        irLed = 1;
        //pinMode(PIN_IR_OUT,OUTPUT);
      }
      uTime += REMOCON_CAPTURE_INTERVAL_US;
      while(timer.read_us() < uTime);
    }
  }
  irLed = 0;
}



void rgbLed(int r, int g, int b) {
    ledR = (float)r/255;
    ledG = (float)g/255;
    ledB = (float)b/255;
}

void offLed () {
    rgbLed(0, 0, 0);    
}

void blinkLed(int r, int g, int b, int dur) {
    rgbLed(r, g, b);
    ledTick.attach(&offLed , dur);
}


// -----------------------------------------------------------------------------
// ----------------------------------- main ------------------------------------
// -----------------------------------------------------------------------------
int main() {  
  while (1) {
    int freq;
      
    switch(state) {
        case START:
            rgbLed(0,0,0);
            timer.start(); // プログラム開始時間を記録するためのtimerスタート
            smpBufp = (double*)malloc( sizeof(double) * FFTSIZE ); //Buffer領域を確保
            dbg("\r\n   --- GuitaRemo Start ---   \r\n"); 
          
            toWait(); // 初期化 + wait遷移処理
            
            break;
        
        
        case WAIT:  // 録音を繰り返しながら演奏開始を待つ   
            freq = detectFreq();
            if(freq != 0) {
                phrase[0] = freq; // フレーズの1音目を登録
                dbg("Record\r\n"); //遷移
                state = RECORD;
            }

            if(sw1) {
              dbg("registration mode \r\n");
              bool result = captureIrCommand(data);
              if(result) {
                for (int i=0; i<3; i++) {
                  rgbLed(255,0,0);
                  wait(0.5);
                  offLed();              
                  wait(0.5);
                  memorize_flag = 1;
                }
              }
            }
            break;
        
        case RECORD: // フレーズ録音
            rgbLed(255,0,255);      
            freq = detectFreq();
            //dbg("%d\r\n",freq); 
            
            if(p_cnt==0){
                phrase[p_cnt] = freq;
                //dbg("%d\r\n",freq); 
            } else { //2音目以降は判定する
                int pre = phrase[p_cnt-1]; //一個前の音のfrequency
            
                if(freq <= 2*pre+FFT_RESOLUTION && freq >= 2*pre-FFT_RESOLUTION ) {//前の音の2倍音なら
                    phrase[p_cnt] = pre; //前の音をバッファに録音する
                } else if (freq <= 3*pre+FFT_RESOLUTION && freq >= 3*pre-FFT_RESOLUTION ) {//前の音の3倍音なら
                    phrase[p_cnt] = pre;
                } else if (freq <= pre+FFT_RESOLUTION && freq >= pre-FFT_RESOLUTION ) {//前の音との差がFFT解像度の範囲なら
                    phrase[p_cnt] = pre;
                } else if (pre <= 2*freq+FFT_RESOLUTION && pre >= 2*freq-FFT_RESOLUTION ) {//前の音が2倍音なら)
                    phrase[p_cnt] = freq;
                    phrase[p_cnt-1] = freq; //前の音も修正
                } else if (pre <= 3*freq+FFT_RESOLUTION && pre >= 3*freq-FFT_RESOLUTION ) {//前の音が3倍音なら)
                    phrase[p_cnt] = freq;
                    phrase[p_cnt-1] = freq; //前の音も修正
                } else {
                    phrase[p_cnt] = freq; //新しい音をバッファに録音する
                }
                // dbg("%d  (freq: %d,  pre: %d\r\n", phrase[p_cnt], freq, pre); 
                //if(phrase[p_cnt] != 0 && phrase[p_cnt] < 550) {
                dbg("%d \r\n", phrase[p_cnt]); 
//                }
                
            }
            p_cnt++;
            
            //dbg("%d\r\n",phrase[p_cnt-1]); 
            
            
            if (p_cnt > 20) { // 20カウント溜まったら
                int recent_sum = 0; 
            //int n = (p_cnt < 20)?p_cnt:20;
                for (int i=0; i<20; i++) { // 直近20カウントの周波数の和を計算
                    recent_sum += phrase[p_cnt-i]; 
                }
                
                // 直近20カウントで周波数の和が0 = 演奏されてない場合
                if( recent_sum == 0 || p_cnt>=PHRASE_LEN) {
                    dbg("Recording Finish\r\n");
                    rgbLed(0,0,0);
                
                    // フレーズを要約(p_cnt: pharaseの配列長, s_cnt: summaryの配列長)
                    s_cnt = 0;
                    for(int i=0; i<p_cnt-1; i++) { // 最後の一音は0なので外す
                        if (i == 0){
                        //最初の1音
                            summary[0][s_cnt++] = phrase[0]; // そのまま周波数を記録
                            soundLength = 0; // 音の長さを初期化
                        } else if ( phrase[i]>phrase[i-1]+FFT_RESOLUTION*2 || phrase[i]<phrase[i-1]-FFT_RESOLUTION*2 ) {
                        //異なったfreqのとき
                            summary[0][s_cnt] = phrase[i]; //新しい音を記録
                            summary[1][s_cnt-1] = soundLength+1; //前の音の長さを記録
                            s_cnt++;
                            soundLength = 0; //　音の長さを初期化
                        } else { 
                        //近いfreqのとき
                            soundLength++; // 音の長さを++
                        }
                    }
                    //dbg("----------\r\n");
                    for(int i=0; i<s_cnt; i++) {
                    //    dbg("freq:%d  length:%d\r\n", summary[0][i], summary[1][i]);            
                    }
                    //dbg("----------\r\n");
                    
                    // state遷移
                    if (memorize_flag) {
                        dbg("MEMORIZE\r\n");
                        state = MEMORIZE;    
                    } else {
                        dbg("CHECK\r\n");
                        state = CHECK;              
                    }
                    
                }
                
            }
            break;
        
        case MEMORIZE:
        
            for (int i=0; i<100; i++) {
                memory[0][i] = 0;
                memory[1][i] = 0;    
            }
            
            for(int i=0; i<s_cnt; i++) {
                memory[0][i] = summary[0][i];
                memory[1][i] = summary[1][i];
            }
            m_len = s_cnt-1;
            memorize_flag = 0;
            toWait();
            
            break;
            
        case CHECK:          
            int s_len = s_cnt-1;
            int n=(m_len>s_len)?m_len:s_len;
            dbg("----------\r\n");
                for(int i=0; i<n; i++) {
                    dbg("memory:%d (%d)  summary:%d (%d)\r\n", memory[0][i], memory[1][i], summary[0][i], summary[1][i]);            
                }
            dbg("----------\r\n");
            
            int ok_cnt=0;
            int ng_cnt=0;
            
            int s_i = 0;
            int m_i = 0;
            
            for(int i=0; i<n; i++) {
                if(summary[0][s_i] == 0) {
                    s_i++;
                }
                if(memory[0][m_i] == 0 ) {
                    m_i++;
                }
                
                int sum = summary[0][s_i]; 
                int mem = memory[0][m_i];
                if(sum!=0 && mem!=0) { // 次も0ならCheckしない
                    if(sum <= mem + FFT_RESOLUTION && sum >= mem - FFT_RESOLUTION) {
                        dbg("%ok  "); 
                        ok_cnt++;   
                    } else if (sum <= mem*2+ FFT_RESOLUTION && sum >= mem*2 - FFT_RESOLUTION){
                        dbg("ok  "); 
                        ok_cnt++;   
                    } else if (sum <= mem*3+ FFT_RESOLUTION && sum >= mem*3 - FFT_RESOLUTION){
                        dbg("ok  "); 
                        ok_cnt++;   
                    } else if (mem <= sum*2+ FFT_RESOLUTION && mem >= sum*3 - FFT_RESOLUTION){
                        dbg("ok  "); 
                        ok_cnt++;   
                    } else if (mem <= sum*3+ FFT_RESOLUTION && mem >= sum*3 - FFT_RESOLUTION){
                        dbg("ok  "); 
                        ok_cnt++;   
                    } else {
                        dbg("ng  ");
                        ng_cnt++;
                    }
                    dbg(": %d, %d\r\n", mem, sum); 
                    
                }
                s_i++; m_i++;
                    
            }
            
            for(int i = 0; i<m_len; i++) {                
                if(summary[0][i] < memory[0][i] + FFT_RESOLUTION*2 && summary[0][i] > memory[0][i] - FFT_RESOLUTION*2) {
                    dbg("ok  "); 
                    ok_cnt++;   
                } else {
                    dbg("ng  ");
                    ng_cnt++;
                }
            }
            
            float ok_rate = (float)ok_cnt/(ok_cnt+ng_cnt) * 100.0f;
            dbg("rate: %f", ok_rate);
            if(ok_rate >= 80) { // SUCCESS
              dbg("       success! \r\n");
                for(int i = 0; i < 3; i++) {
                  //rgbLed(255, 0, 0);
//                  transmitIrCommand(data); 
//                  wait(0.3);
//                  offLed();
//                  wait(0.5);
                    blinkLed(255,0,0,1);
                }
            } else { // FAIL
                dbg("       fail... \r\n");
                blinkLed(0, 255, 0, 1);
            }
            
            toWait();
            
            break;

        case TEST:    
                    
        break;
    }
    
    
    
     
    
  }
  
  if (smpBufp) { // 終了時、Bufferを初期化?
    free(smpBufp);
  }
  return 0;
}


